// $Id: file.cxx 3779 2005-07-01 10:48:25Z jz $

#include "zschimmer.h"
#include "file.h"
#include "log.h"

#include <sys/stat.h>

#ifdef Z_WINDOWS
#   include <io.h>
#   include <share.h>
#   include "z_windows.h"
#else
#   include <stdio.h>
#   include <unistd.h>
#   include <sys/mman.h>
#   include <pwd.h>
//#   include "z_unix.h"
#endif

#include <errno.h>


#ifndef _MSC_VER
    const int O_BINARY = 0;
#endif

namespace zschimmer {

//--------------------------------------------------------------------------------------------const

static Error_code_text error_codes[] =
{
    { "Z-FILE-101", "open mode ist nicht korrekt angegeben: $1" },
    { "Z-FILE-102", "Datei ist schon geffnet (doppeltes open)" },
    { "Z-FILE-103", "Dateiposition >= 2GB wird auf diesem Betriebsystem nicht untersttzt" },
    { "Z-FILE-104", "Dateigre >= 2GB wird nicht untersttzt" },
    { "Z-FILE-105", "Datei ist geschlossen" },
    { NULL         , NULL }
};

//-------------------------------------------------------------------------------------------Z_INIT

Z_INIT( file )
{
    add_error_code_texts( error_codes );
}

//---------------------------------------------------------------------------------File_base::print

void File_base::print( int i )
{
    char buffer [ 50 ];
    _last_errno = 0;

    int len = sprintf( buffer, "%d", i );
    write( buffer, len );
}

//---------------------------------------------------------------------------File_base::read_string

string File_base::read_string( size_t size )
{
    struct Buffer
    {
        ~Buffer()  { delete _buffer; }
        char* _buffer;
    };
    
    Buffer buffer;
    buffer._buffer = new char[ size ];

    size_t length = read( buffer._buffer, size );
    return string( buffer._buffer, length );
}

//---------------------------------------------------------------------------------------File::File

File::File( const string& filename, const string& mode )
:
    _file(-1),
    _do_unlink(false),
    _is_my_file(false)
{
    open( filename, mode );
}

//--------------------------------------------------------------------------------------File::~File

File::~File()
{
    try
    {
        if( _is_my_file )  close();
    }
    catch( const exception& x ) 
    {
        Z_LOG( "~File(\"" << _filename << "\"): " << x << "\n" );
    }

    if( _do_unlink )  
    {
        Z_LOG( "unlink(\"" << _filename << "\")\n" );
        unlink( _filename.c_str() );
    }
}

//---------------------------------------------------------------------------------------File::open

void File::open( const string& filename, const string& mode, int rights )
{
    int         imode = O_NOINHERIT;
    const char* p     = mode.c_str();

    while( *p )
    {
        switch( *p )
        {
            case 'r': 
            {
                if( *++p == '+' )  imode |= O_RDWR,  p++;
                             else  imode |= O_RDONLY;
                break;
            }

            case 'w': 
            {
                if( *++p == '+' )  imode |= O_TRUNC | O_CREAT | O_RDWR,  p++;
                             else  imode |= O_TRUNC | O_CREAT | O_WRONLY;
                break;
            }

            case 'a': 
            {
                if( *++p == '+' )  imode |= O_APPEND | O_CREAT | O_RDWR,  p++;
                             else  imode |= O_APPEND | O_CREAT | O_WRONLY;
                break;
            }

            case 'b':
            case 'B':
            {
#               ifdef Z_WINDOWS
                    imode |= O_BINARY;
#               endif

                p++;
                break;
            }

            case 'I':   // Inherit
            {
                imode &= ~O_NOINHERIT;
                p++;
                break;
            }

            case 'S':   // Sequential
            {
                Z_WINDOWS_ONLY( imode &= ~_O_SEQUENTIAL );
                p++;
                break;
            }
            
            case 'R':   // Random access
            {
                Z_WINDOWS_ONLY( imode &= ~_O_RANDOM );
                p++;
                break;
            }
            


            default: throw_xc( "Z-FILE-101", mode );
        }
    }

    open( filename, imode, rights );
}

//---------------------------------------------------------------------------------------File::open

void File::open( const string& filename, int mode, int rights )
{
    _last_errno = 0;

    if( opened() )  throw_xc( "Z-FILE-102" );

    _filename = filename;
    _is_my_file = true;

    Z_LOG( "open(\"" << filename << "\")\n" );
    _file = ::open( filename.c_str(), mode, rights );
    if( _file == -1 )  _last_errno = errno, check_error( "open" );
}

//-----------------------------------------------------------------------------File::open_temporary

void File::open_temporary( int flags, string a_name )
{
    _last_errno = 0;

    if( opened() )  throw_xc( "Z-FILE-102" );

    int pmode = S_IREAD | S_IWRITE;


#   ifdef Z_WINDOWS

        create_temporary( 0, a_name );
        
        int oflag = _O_SHORT_LIVED;
        if(    flags & open_unlink        )  oflag |= O_TEMPORARY;      // close() lscht die Datei
        if( !( flags & open_inheritable ) )  oflag |= O_NOINHERIT;      // Nicht an Prozesse vererben (machen wir zum Default, nur Windows)

        Z_LOG( "sopen(\"" << _filename << "\")\n" );
        _file = ::sopen( _filename.c_str(), oflag | O_CREAT | O_TRUNC | O_RDWR | O_BINARY, flags & open_private? _SH_DENYRW : _SH_DENYWR, pmode );
        if( _file == -1 )  _last_errno = errno, throw_errno( _last_errno, "open", _filename.c_str() );

#    else

#       ifdef Z_LINUX

            if( strchr( a_name.c_str(), '/' ) )  _filename = a_name;
                                           else  _filename = get_temp_path() + "/" + a_name;

            _filename += ".XXXXXX";

            Z_LOG( "mkstemp(\"" << _filename << "\")  " );
            _file = mkstemp( (char*)_filename.c_str() );        // ffnet die Datei
            if( _file == -1 )  throw_errno( errno, "mkstemp", _filename.c_str() );
            Z_LOG( "=> " << _filename << "\n" );

#        else

            string directory = directory_of_path( a_name );
            string prefix    = filename_of_path( a_name );

            if( directory.empty() )  directory = get_temp_path();

            for( int i = 0; i < 100; i++ )
            {
                Z_LOG( "tempnam(\"" << directory << "\",\"" << prefix << "\")  " );

                char* filename_ptr = tempnam( directory.empty()? NULL : directory.c_str(), prefix.c_str() );
                if( !filename_ptr )  _last_errno = errno, throw_errno( _last_errno, "tempnam" );

                _filename = filename_ptr;
                free( filename_ptr );

                Z_LOG( "=> " << _filename << "\n" );

                Z_LOG( "open(\"" << _filename << "\")\n" );
                _file = ::open( _filename.c_str(), O_EXCL | O_CREAT | O_RDWR, pmode );
                if( _file == -1 )
                {
                    if( errno == EEXIST )  continue;
                    _last_errno = errno, throw_errno( _last_errno, "open", _filename.c_str() );
                }

                break;
            }

#       endif

        if( flags & open_unlink )
        {
            Z_LOG( "unlink(\"" << _filename << "\")\n" );
            int ret = unlink( _filename.c_str() );                    // Namen sofort wieder lschen
            if( ret == -1 )  _last_errno = errno, throw_errno( _last_errno, "unlink", _filename.c_str() );
        }

#   endif

    _do_unlink = ( flags & open_unlink_later ) != 0;
}

//-----------------------------------------------------------------------------File::open_temporary

void File::create_temporary( int flags, string a_name )
{
    _last_errno = 0;

    if( opened() )  throw_xc( "Z-FILE-102" );


#   ifdef Z_WINDOWS

        char tmp_filename [ MAX_PATH ];

        string path = get_temp_path();
        int ok = GetTempFileName( path.c_str(), a_name.c_str(), 0, tmp_filename );
        if( !ok )  throw_mswin( "GetTempPath" );

        _filename = tmp_filename;

#    else

        open_temporary( 0, a_name );
        close();

#   endif

    _do_unlink = ( flags & open_unlink_later ) != 0;
}

//--------------------------------------------------------------------------------------File::close

void File::close()
{
    _last_errno = 0;

    if( _file == -1 )  return;

    int ret = ::close( _file );
    _file = -1;

    if( ret == -1 )  _last_errno = errno, check_error( "close" );
}

//------------------------------------------------------------------------------File::assign_fileno

void File::assign_fileno( int f )
{
    if( opened() )  throw_xc( "Z-FILE-102" );

    _file = f;
    _is_my_file = false;
}

//--------------------------------------------------------------------------------------File::write

void File::write( const void* p, size_t len )
{
    _last_errno = 0;

    size_t ret = ::write( _file,  p, len );
    if( ret != len )  _last_errno = errno, check_error( "write" );
}

//---------------------------------------------------------------------------------------File::read

size_t File::read( void* p, size_t len )
{
    _last_errno = 0;

    size_t ret = ::read( _file, p, len );

    if( ret == (size_t)-1 )  _last_errno = errno, check_error( "read" );
    return ret;
}

//---------------------------------------------------------------------------------------File::seek

void File::seek( int64 pos, int origin )
{
    _last_errno = 0;

#   ifdef Z_WINDOWS
        int64 ret = _lseeki64( _file, pos, origin );
#    else
        if( pos >> 31 )  throw_xc( "Z-FILE-103" );
        int ret = lseek( _file, pos, origin );
#   endif

    if( ret == -1 )  _last_errno = errno, check_error( "seek" );
}

//---------------------------------------------------------------------------------------File::tell

int64 File::tell()
{
    _last_errno = 0;

#   ifdef Z_WINDOWS
        int64 ret = _telli64( _file );
        if( ret == -1 )  _last_errno = errno, check_error( "_telli64" );
        return ret;
#    else
        int ret = lseek( _file, 0, SEEK_CUR );
        if( ret == -1 )  _last_errno = errno, check_error( "lseek" );
        return ret;
#   endif

}

//--------------------------------------------------------------------------------------File::flush

void File::flush()
{
    // Nicht zu tun
}

//---------------------------------------------------------------------------------------File::sync

void File::sync()
{
    _last_errno = 0;

    flush();

#   ifdef Z_WINDOWS
        int ret = _commit( _file );
        if( ret == -1 )  _last_errno = errno, check_error( "_commit" );
#    else
        int ret = fsync( _file );
        if( ret == -1 )  _last_errno = errno, check_error( "fsync" );
#   endif
}

//-----------------------------------------------------------------------------------File::syncdata       

void File::syncdata()
{
#   if defined Z_WINDOWS || defined Z_SOLARIS || defined Z_HPUX

        sync();

#    else

        _last_errno = 0;

        int ret = fdatasync( _file );
        if( ret == -1 )  _last_errno = errno, check_error( "fdatasync" );

#   endif
}

//--------------------------------------------------------------------------------File::check_error

void File::check_error( const char* operation )
{
    if( _last_errno )  throw_errno( _last_errno, _filename.c_str(), operation );
}

//-------------------------------------------------------------------------------------File::handle
#ifdef Z_WINDOWS

HANDLE File::handle() const
{ 
    return (HANDLE)_get_osfhandle( _file ); 
}

#endif
//-------------------------------------------------------------------------Mapped_file::Mapped_file

Mapped_file::Mapped_file( const string& filename, const string& mode ) 
: 
    _zero_(this+1)
{
     open( filename, mode );
}

//------------------------------------------------------------------------Mapped_file::~Mapped_file

Mapped_file::~Mapped_file()
{
#   ifdef Z_WINDOWS

        UnmapViewOfFile( _ptr );

#    else

        if( _ptr && _ptr != MAP_FAILED )  ::munmap( _ptr, 0 );

#   endif 
}

//--------------------------------------------------------------------------------Mapped_file::open

void Mapped_file::open( const string& filename, const string& mode, int rights ) 
{ 
    File::open( filename, mode, rights ); 

    //map(); 
}

//--------------------------------------------------------------------------------Mapped_file::open

void Mapped_file::open( const string& filename, int mode, int rights ) 
{ 
    File::open( filename, mode, rights ); 

    //map(); 
}

//---------------------------------------------------------------------------Mapped_file::as_string

string Mapped_file::as_string()
{ 
    return string( (const char*)map(), map_length() ); 
}

//---------------------------------------------------------------------------------Mapped_file::map

void* Mapped_file::map()
{
#   ifdef Z_WINDOWS

        if( !_ptr && !_map )
        {
            int64 length64 = this->length();
            if( length64 > SIZE_MAX )  throw_xc( "Z-FILE-104", length64 );
            
            size_t length = (size_t)length64;

            if( length == -1 )  _last_errno = errno, throw_errno( _last_errno, "filelength", _filename.c_str() );

            if( length > 0 )
            {
                _map = CreateFileMapping( (void*)_get_osfhandle( _file ), NULL, PAGE_READONLY, 0, length, NULL );           
                if( !_map )  throw_mswin( "CreateFileMapping", _filename.c_str() );

                _ptr = MapViewOfFile( _map, FILE_MAP_READ, 0, 0, length );
                if( !_ptr )  throw_mswin( "MapViewOfFile", _filename.c_str() );
            }

            _map_length = length;
        }

#    else
    
        if( _ptr == NULL || _ptr == MAP_FAILED )
        {
            int64 length64 = this->length();
            if( length64 > SIZE_MAX )  throw_xc( "Z-FILE-104", length64 );
            
            size_t length = (size_t)length64;

            if( length > 0 )
            {
                _ptr = mmap( NULL, length, PROT_READ, MAP_PRIVATE, _file, 0 );
                if( _ptr == MAP_FAILED )  _last_errno = errno, throw_errno( _last_errno, "mmap", _filename.c_str() );
            }

            _map_length = length;
        }

#   endif 

    return _ptr;
}

//-------------------------------------------------------------------------------Mapped_file::unmap

void Mapped_file::unmap()
{
#   ifdef Z_WINDOWS

        if( _ptr )  
        {
            BOOL ok = UnmapViewOfFile( _ptr );
            _ptr = NULL;
            if( !ok )  throw_mswin( "UnmapViewOfFile", _filename.c_str() );

            _map.close();
        }

#    else

        if( _ptr && _ptr != MAP_FAILED )
        {
            int err = ::munmap( _ptr, _map_length );
            _ptr = NULL;
            if( err )  _last_errno = errno, throw_errno( _last_errno, "munmap", _filename.c_str() );
        }

#   endif 

    _map_length = 0;
}

//-------------------------------------------------------------------------------------------------

//------------------------------------------------------------------------Stream_file::~Stream_file

Stream_file::~Stream_file()
{
    try
    {
        if( _is_my_file )
        {
            close();

            if( _do_unlink )  
            {
                Z_LOG( "unlink(\"" << _filename << "\")\n" );
                unlink( _filename.c_str() );
            }
        }
    }
    catch( const exception& ) {}
}

//--------------------------------------------------------------------------------Stream_file::open

void Stream_file::open( const string& filename, const string& mode )
{
    bool ok = try_open( filename, mode );
    if( !ok )  check_error( "open" );
}

//----------------------------------------------------------------------------Stream_file::try_open

bool Stream_file::try_open( const string& filename, const string& mode )
{
    _last_errno = 0;

    if( opened() )  throw_xc( "Z-FILE-102" );

    _filename = filename;
    
    _is_my_file = true;
    _file = fopen( filename.c_str(), mode.c_str() );


    if( _file == NULL )
    {
        _last_errno = errno;
        return false;
    }
    else
        return true;
}

//-------------------------------------------------------------------------------Stream_file::close

void Stream_file::close()
{
    _last_errno = 0;

    if( !_file )  return;

    int ret = fclose( _file );
    _file = NULL;

    if( ret == -1 )  check_error( "fclose" );
}

//-------------------------------------------------------------------------------Stream_file::write

void Stream_file::write( const void* p, size_t len )
{
    _last_errno = 0;

    size_t ret = fwrite( p, 1, len, _file );
    if( ret != len )  check_error( "fwrite" );
}

//-----------------------------------------------------------------------------Stream_file::put_char

void Stream_file::put_char( char c )
{
    _last_errno = 0;

    int ret = fputc( c, _file );
    if( ret == -1 )  check_error( "fputc" );
}

//-------------------------------------------------------------------------------Stream_file::write
/*
void Stream_file::printf( const char* format, ... )
{
    _last_errno = 0;

    int ret = fprintf( _file, format, ... );
    if( ret != -1 )  check_error();
}
*/
//--------------------------------------------------------------------------------Stream_file::read

size_t Stream_file::read( void* p, size_t len )
{
    _last_errno = 0;

    size_t ret = fread( p, 1, len, _file );
    if( ret == (size_t)-1 )  check_error( "fread" );
    return ret;
}

//--------------------------------------------------------------------------------Stream_file::seek

void Stream_file::seek( int64 pos, int origin )
{
    _last_errno = 0;

    if( pos > LONG_MAX )  throw_xc( "Z-FILE-103" );

    int ret = fseek( _file, (long)pos, origin );
    if( ret == -1 )  check_error( "fseek" );
}

//--------------------------------------------------------------------------------Stream_file::tell

int64 Stream_file::tell()
{
    _last_errno = 0;

    int ret = ftell( _file );
    if( ret == -1 )  check_error( "ftell" );
    return ret;
}

//-------------------------------------------------------------------------------Stream_file::flush

void Stream_file::flush()
{
    _last_errno = 0;

    int ret = fflush( _file );
    if( ret == -1 )  check_error( "fflush" );
}

//--------------------------------------------------------------------------------Stream_file::sync

void Stream_file::sync()
{
    _last_errno = 0;

    flush();

#   ifdef Z_WINDOWS
        int ret = _commit( fileno( _file ) );
        if( ret == -1 )  check_error( "_commit" );
#    else
        int ret = fsync( fileno( _file ) );
        if( ret == -1 )  check_error( "fsync" );
#   endif

}

//----------------------------------------------------------------------------Stream_file::syncdata       

void Stream_file::syncdata()
{
#   if defined Z_WINDOWS || defined Z_SOLARIS || defined Z_HPUX

        sync();

#    else

        _last_errno = 0;

        int ret = fflush( _file );
        if( ret == -1 )  check_error( "fflush" );

        ret = fdatasync( fileno( _file ) );
        if( ret == -1 )  check_error( "fdatasync" );

#   endif
}

//-------------------------------------------------------------------------Stream_file::check_error

void Stream_file::check_error( const char* operation )
{
    if( _last_errno )  throw_errno( _last_errno, _filename.c_str(), operation );
}

//-------------------------------------------------------------------------------------------------
/*
    Fr eine Klasse, die zeilenweise liest:
    (Line_reader, oder wie heit die entsprechende Klasse in Java?)

    string line;
    line.reserve( 1000 );

    for( double until = double_from_gmtime() + timeout; until > double_from_gmtime(); )
    {
        string part = vmware_log.read_string( 200 );

        if( part == "" )
        {
            my_log << '.';
            _controller->my_sleep( 0.1 );
        }
        else
        {
            while( part != "" )
            {
                size_t n = part.find( '\n' );

                if( n == string::npos )
                {
                    line += part;
                    part = "";
                }
                else
                {
                    line.append( part.data(), n );
                    part.erase( 0, n + 1 );

                    line.erase( 0, 17 );            // "May 11 12:44:58: "

                    if( regex.match( line ) ) 
                    {
                        my_log << " ok\n";
                        return true;
                    }

                    line = "";
                }
            }
        }
    }
*/
//---------------------------------------------------------------------------------string_from_file

string string_from_file( const string& filename )
{ 
    Mapped_file file ( filename, "r" );

    return string_from_mapped_file( &file );
}

//---------------------------------------------------------------------------------string_from_file

string string_from_fileno( int fileno )
{ 
    Mapped_file file;
    
    file.assign_fileno( fileno );

    return string_from_mapped_file( &file );
}

//--------------------------------------------------------------------------string_from_mapped_file

string string_from_mapped_file( Mapped_file* file )
{
    if( !file->opened() )  throw_xc( "Z-FILE-105", file->filename() );

    off_t ret = lseek( file->_file, 0, SEEK_SET );
    if( ret == (off_t)-1 )
    {
        string text;
        int    allocate = 10*1024;

        while(1)
        {
            char   buffer [ 16*1024 ];
            size_t length = file->read( &buffer, sizeof buffer );
            if( length == 0 )  break;

            if( text.capacity() < text.length() + length )
            {
                if( allocate < 100*1024*1024 )  allocate = 2*allocate;
                text.reserve( text.length() + allocate );
            }

            text.append( buffer, length );
        }

        return text;
    }
    else
    {
        return file->as_string();
    }
}

//------------------------------------------------------------------------modification_time_of_file

time_t modification_time_of_file( const string& filename )
{
    struct stat s;
    int err = ::stat( filename.c_str(), &s );                  
    if( err ) throw_errno( errno, filename.c_str() );

    return s.st_mtime;                      // Ist lokale Zeit bei lokalen Dateien, GMT bei Samba (Fehler in Samba?)
}

//-------------------------------------------------------------------------------------z_filelength

int64 z_filelength( const char* filename )
{
    struct stat s;

    int err = ::stat( filename, &s );
    if( err ) throw_errno( errno, filename );

    return s.st_size;
}

//-------------------------------------------------------------------------------------z_filelength

int64 z_filelength( int file_handle )
{
#   ifdef Z_WINDOWS

        int64 result = ::filelength( file_handle );
        if( result == -1 )  throw_errno( errno, "filelength" );
        return result;

#    else

        struct stat s;

        int err = ::fstat( file_handle, &s );
        if( err ) throw_errno( errno );

        return s.st_size;

#   endif
}

//-----------------------------------------------------------------------------------------z_unlink

void z_unlink( const char* filename )
{
    int err = unlink( filename );
    if( err ) throw_errno( errno, filename );
}

//--------------------------------------------------------------------------------get_temp_path

string get_temp_path()
{
#   ifdef Z_WINDOWS

        char buffer [_MAX_PATH+1];

        int len = GetTempPath( sizeof buffer, buffer );
        if( len == 0 )  throw_mswin( "GetTempPath" );

        return string( buffer, len );

#    else

        char* t = getenv( "TMP" );
        string tmp = t? t : "";

        if( tmp == "" ||  tmp == "/tmp" )  
        {
            tmp = "/tmp/";
            int uid = getuid();
            struct passwd* p = getpwuid( uid );
            if( p )  tmp += p->pw_name;
               else  tmp += as_string(uid);

            mkdir( tmp.c_str(), (uint)-1 );       // Fehler ignorieren wir
        }

        return tmp;

#   endif
}

//--------------------------------------------------------------------------------------file_exists

bool file_exists( const string& filename )
{
#   ifdef Z_WINDOWS    

        return GetFileAttributes( filename.c_str() ) != -1;

#    else

        struct stat s;

        int err = ::stat( filename.c_str(), &s );

        return !err;

#   endif
}

//----------------------------------------------------------------------------directory_of_path
// Liefert alles auer dem Datennamen und Schrgstrich

string directory_of_path( const string& path )
{
    const char* p0 = c_str( path );
    const char* p  = p0 + length( path );

    if( p > p0 )
    {
#       if defined Z_WINDOWS
            if( length(path) >= 2  &&  p0[1] == ':' )  p0 += 2;
            while( p > p0  &&  p[-1] != '/'  &&  p[-1] != '\\'  &&  p[-1] != ':' )  p--;
            if( p > p0+1  &&  ( p[-1] == '/' || p[-1] == '\\' ) )  p--;
#        else
            while( p > p0  &&  p[-1] != '/' )  p--;
            if( p > p0+1  &&  p[-1] == '/' )  p--;
#       endif
    }

    return path.substr( 0, p - c_str(path) );
}

//-----------------------------------------------------------------------------filename_of_path
// Liefert den Dateinamen von "pfad/dateiname"

string filename_of_path( const string& path )
{
    const char* p0 = path.c_str();
    const char* p  = p0 + path.length();

#   if defined Z_WINDOWS
        while( p > p0  &&  p[-1] != ':'  &&  p[-1] != '/'  &&  p[-1] != '\\' )  p--;
#    else
        while( p > p0  &&  p[-1] != '/' )  p--;
#   endif

    return p;
}

//-----------------------------------------------------------------------------basename_of_path
// Liefert den Basisnamen  "pfad/basisname.ext"

string basename_of_path( const string& path )
{
    const char* p0 = c_str( path );
    const char* p2 = p0 + length( path );
    const char* p1;

#   if defined Z_WINDOWS
        while( p2 > p0  &&  p2[-1] != '.'  &&  p2[-1] != '/'  &&  p2[-1] != '\\' )  p2--;
        while( p2 > p0  &&  p2[-1] == '.' )  p2--;
        p1 = p2;
        while( p1 > p0  &&  p1[-1] != '/'  &&  p1[-1] != '\\' )  p1--;
#   else
        while( p2 > p0  &&  p2[-1] != '.'  &&  p2[-1] != '/' )  p2--;
        while( p2 > p0  &&  p2[-1] == '.' )  p2--;
        p1 = p2;
        while( p1 > p0  &&  p1[-1] != '/' )  p1--;
#   endif

    return string( p1, p2 - p1 );
}

//----------------------------------------------------------------------------extension_of_path
// Liefert die Extension "pfad/basisname.ext"

string extension_of_path( const string& path )
{
    const char* p0 = c_str( path );
    const char* p2 = p0 + length( path );

    while( p2 > p0  &&  p2[-1] == ' '  )  p2--;
    const char* p1 = p2;

#   if defined Z_WINDOWS
        while( p1 > p0  &&  p1[-1] != '.'  &&  p1[-1] != '/'  &&  p1[-1] != '\\' )  p1--;
#    else
        while( p1 > p0  &&  p1[-1] != '.'  &&  p1[-1] != '/' )  p1--;
#   endif

    if( p1[-1] != '.' )  p1 = p2;
    return string( p1, p2 - p1 );
}

//-------------------------------------------------------------------------------------------------

} //zschimmer
